! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: linear
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains some routines to do with least-squares estimation and solution of linear equations.
! 
! Comments: TODO:
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module linear

  use constants

  implicit none 

logical, parameter :: testFlag = .false.

  private 

  public solveLinearTriDiag

contains 

  !---------------------------------------------------------------------
  subroutine solveLinearTriDiagFisher(triDiag, b, x, status)
  !---------------------------------------------------------------------
  ! This version was written from the algorithm in M E Fisher. It's the
  ! same as the other, just more explicit. We'll keep it just for reference.

    implicit none

    real, intent(in) :: b(:), triDiag(-1:1, size(b))
    real, intent(out) :: x(size(b))
    integer, intent(out), optional :: status

    integer :: localStatus, n, j
    real :: decomp(-1:1, size(b)), tempB(size(b))

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    n = size(b)
  if (n <= 0) return ! trivial.

    x = 0.0

    ! LU decompose:
    decomp = 0.0
    decomp(1,:n-1) = triDiag(1,:n-1)
    decomp(0,1) = triDiag(0,1)
    do j = 2, n
      decomp(-1,j) = triDiag(-1,j) / decomp(0,j-1)
      decomp(0,j) = triDiag(0,j) - decomp(-1,j) * decomp(1,j-1)
    end do

    ! Forward substitution:
    tempB(1) = b(1)
    do j = 2, n
      tempB(j) = b(j) - decomp(-1,j) * tempB(j-1)
    end do

    ! Backward substitution:
    x(n) = tempB(n) / decomp(0,n)
    do j = n - 1, 1, -1
      x(j) = (tempB(j) - decomp(1,j) * x(j+1)) / decomp(0,j)
    end do

    if (present(status)) status = localStatus

  end subroutine solveLinearTriDiagFisher

  !---------------------------------------------------------------------
  subroutine solveLinearTriDiag(triDiag, b, x, status)
  !---------------------------------------------------------------------
  ! Adapted from Numerical Recipes for Fortran.

    implicit none

    real, intent(in) :: b(:), triDiag(-1:1, size(b))
    real, intent(out) :: x(size(b))
    integer, intent(out), optional :: status

    integer :: localStatus, n, j
    real :: bet, gam(size(b))

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    n = size(b)
  if (n <= 0) return ! trivial.

    x = 0.0

    bet = triDiag(0, 1)
    x(1) = b(1) / bet
    do j = 2, n
      gam(j) = triDiag(1, j-1) / bet
      bet = triDiag(0, j) - triDiag(-1, j) * gam(j)

      if (bet == 0.0) then
        if (present(status)) status = -1
  return
      end if

      x(j) = (b(j) - triDiag(-1, j) * x(j-1)) / bet
    end do
    do j = n - 1, 1, -1
      x(j) = x(j) - gam(j+1) * x(j+1)
    end do

    if (present(status)) status = localStatus

  end subroutine solveLinearTriDiag

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module linear
