! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module sort_mod
  !  Non-recursive stack versions of quicksort from N.Wirth's Pascal
  !  book, "Algorithms + data structures = programs".
  !  Converted to Fortran90 by Alan Miller, CSIRO.
  !  Module constructed by Clive Page, Leicester University.
  implicit none

  interface quick_sort
     module procedure &
          quick_sort_c,  quick_sort_d,  quick_sort_i,  quick_sort_r, &
          quick_sort_co, quick_sort_do, quick_sort_io, quick_sort_ro
  end interface

contains

  subroutine quick_sort_c(array)
    character(*), intent(INOUT)    :: array(:)
    character(LEN=len(array)) :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(30), stackr(30)

    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit  
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                array(i) = array(j)
                array(j) = temp
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
    return
  end subroutine quick_sort_c

  subroutine quick_sort_d(array)
    double precision, intent(INOUT)    :: array(:)
    double precision :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(30), stackr(30)

    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit  
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                array(i) = array(j)
                array(j) = temp
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
    return
  end subroutine quick_sort_d

  subroutine quick_sort_i(array)
    integer, intent(INOUT)    :: array(:)
    integer :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(30), stackr(30)

    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit  
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                array(i) = array(j)
                array(j) = temp
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
    return
  end subroutine quick_sort_i

  subroutine quick_sort_r(array)
    real, intent(INOUT)    :: array(:)
    real     :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(30), stackr(30)

    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit  
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                array(i) = array(j)
                array(j) = temp
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
    return
  end subroutine quick_sort_r

  subroutine quick_sort_co(array, order)
    character(*), intent(INOUT)    :: array(:)
    integer, intent(OUT) :: order(:)  ! returns original position
    character(LEN=len(array)) :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(15), stackr(15), ww

    do j = lbound(array,1),ubound(array,1)
       order(j) = j
    end do
    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                ww   = order(i)
                array(i) = array(j)
                order(i) = order(j)
                array(j) = temp
                order(j) = ww
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
    return
  end subroutine quick_sort_co

  subroutine quick_sort_do(array, order)
    double precision, intent(INOUT)    :: array(:)
    integer, intent(OUT) :: order(:)  ! returns original position
    double precision :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(15), stackr(15), ww

    do j = lbound(array,1),ubound(array,1)
       order(j) = j
    end do
    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                ww   = order(i)
                array(i) = array(j)
                order(i) = order(j)
                array(j) = temp
                order(j) = ww
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
    return
  end subroutine quick_sort_do

  subroutine quick_sort_io(array, order)
    integer, intent(INOUT)    :: array(:)
    integer, intent(OUT) :: order(:)  ! returns original position
    integer  :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(15), stackr(15), ww

    do j = lbound(array,1),ubound(array,1)
       order(j) = j
    end do
    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                ww   = order(i)
                array(i) = array(j)
                order(i) = order(j)
                array(j) = temp
                order(j) = ww
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
    return
  end subroutine quick_sort_io

  subroutine quick_sort_ro(array, order)
    real, intent(INOUT)    :: array(:)
    integer, intent(OUT) :: order(:)  ! returns original position
    real     :: temp, x

    !     Local Variables

    integer  :: i, j, k, l, r, s, stackl(15), stackr(15), ww

    do j = lbound(array,1),ubound(array,1)
       order(j) = j
    end do
    s = lbound(array,1)
    stackl(1) = lbound(array,1)
    stackr(1) = ubound(array,1)

    !     KEEP TAKING THE TOP REQUEST FROM THE STACK UNTIL S = 0.

    do
       l = stackl(s)
       r = stackr(s)
       s = s - 1

       !     KEEP SPLITTING array(L), ... , array(R) UNTIL L >= R.

       do
          i = l
          j = r
          k = (l+r) / 2
          x = array(k)

          !     REPEAT UNTIL I > J.

          do
             do                          ! Search from lower end
                if (array(i) >= x) exit
                i = i + 1
             end do

             do                          ! Search from upper end
                if( x >= array(j)) exit
                j = j - 1
             end do

             if (i <= j) then                     ! Swap positions i & j
                temp    = array(i)
                ww   = order(i)
                array(i) = array(j)
                order(i) = order(j)
                array(j) = temp
                order(j) = ww
                i = i + 1
                j = j - 1
                if (i > j) exit
             else
                exit
             end if
          end do

          if (j-l >= r-i) then
             if (l < j) then
                s = s + 1
                stackl(s) = l
                stackr(s) = j
             end if
             l = i
          else
             if (i < r) then
                s = s + 1
                stackl(s) = i
                stackr(s) = r
             end if
             r = j
          end if
          if(l >= r) exit
       end do
       if (s == 0) exit
    end do
    return
  end subroutine quick_sort_ro

end module sort_mod
