! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: poissontool_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This task allows the user to make spot calculations of several higher mathematical functions connected with the Poisson distribution.
! 
! Comments: 
!
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module poissontool_mod

  use dal
  use param  
  use errorHandling

  use math_utils
  use dal_aux
  use detection_stats

  implicit none 

  private 

  ! Command-line parameters:

  public poissontool

contains 

  !---------------------------------------------------------------------
  subroutine poissontool
  !---------------------------------------------------------------------

    implicit none

    real :: realValue, myAnswer, myAnswerUncert, aValue, xValue, bkgValue, likeValue
    integer :: intValue, row
    character(255) :: setTabName, setName, tabName
    character(20) :: realColName, intColName, answerColName, aColName, xColName, bkgColName, likeColName
    type(DataSetT) :: set
    type(TableT) :: tab
!    type(ColumnT) :: 
    real(single), pointer :: realValues(:), answerValues(:)
    real(single), pointer :: aValues(:), xValues(:), bkgValues(:), likeValues(:)
    integer(int32), pointer :: intValues(:)
    integer :: foreignStatus

    !-------------------------------------------------------------------
    ! Start

!    call readParameters

    select case(stringParameter('function'))
    case('poisson')
      if (stringParameter('style') .eq. 'command') then
        realValue = realParameter('reals', 0)
        intValue = intParameter('ints', 0)
        myAnswer = poissonProb(realValue, intValue)
        print *, 'answer = ', myAnswer

      else
        setTabName = stringParameter('tab')
        realColName = stringParameter('realcols', 0)
        intColName = stringParameter('intcols', 0)
        answerColName = stringParameter('answercol')

        call splitSetTabName(setTabName, setName, tabName)

        set = dataSet(setName, MODIFY)
        tab = table(set, tabName)
        realValues => real32Data(column(tab, realColName))
        intValues  => int32Data( column(tab, intColName))

        if (hasColumn(tab, answerColName)) then
          call deleteColumn(tab, answerColName)
        end if
        answerValues => real32Data(addColumn(tab, answerColName, REAL32))

        do row = 0, numberOfRows(tab) - 1
          answerValues(row) = poissonProb(realValues(row), intValues(row))
        end do

        call release(set)
      end if

    case('intcpoisson')
      if (stringParameter('style') .eq. 'command') then
        realValue = realParameter('reals', 0)
        intValue = intParameter('ints', 0)
        myAnswer = integratedPoissonProb(realValue, intValue)
        print *, 'answer = ', myAnswer

      else
        setTabName = stringParameter('tab')
        realColName = stringParameter('realcols', 0)
        intColName = stringParameter('intcols', 0)
        answerColName = stringParameter('answercol')

        call splitSetTabName(setTabName, setName, tabName)

        set = dataSet(setName, MODIFY)
        tab = table(set, tabName)
!        realValues => real32Data(column(tab, realColName))
!        intValues  => int32Data( column(tab, intColName))
call readColDataToPtr(tab, realColName, realValues)
call readColDataToPtr(tab, intColName,  intValues)

!        if (hasColumn(tab, answerColName)) then
!          call deleteColumn(tab, answerColName)
!        end if
!        answerValues => real32Data(addColumn(tab, answerColName, REAL32))
answerValues => real32Data(addOrOpenColumn(tab, answerColName, REAL32))

        do row = 0, numberOfRows(tab) - 1
          answerValues(row) = integratedPoissonProb(realValues(row), intValues(row))
        end do

        call release(set)
      end if

    case('incompletegammap')
      if (stringParameter('style') .eq. 'command') then
        aValue = realParameter('reals', 0)
        xValue = realParameter('reals', 1)
        call incompleteGammaP(aValue, xValue, myAnswer, foreignStatus)
if (foreignStatus .ne. 0) call error('badArguments','')
        print *, 'answer = ', myAnswer

      else
        setTabName = stringParameter('tab')
        aColName = stringParameter('realcols', 0)
        xColName = stringParameter('realcols', 1)
        answerColName = stringParameter('answercol')

        call splitSetTabName(setTabName, setName, tabName)

        set = dataSet(setName, MODIFY)
        tab = table(set, tabName)
        aValues => real32Data(column(tab, aColName))
        xValues => real32Data(column(tab, xColName))

        if (hasColumn(tab, answerColName)) then
          call deleteColumn(tab, answerColName)
        end if
        answerValues => real32Data(addColumn(tab, answerColName, REAL32))

        do row = 0, numberOfRows(tab) - 1
          call incompleteGammaP(aValues(row), xValues(row), answerValues(row)&
            , foreignStatus)
if (foreignStatus .ne. 0) call error('badArguments','')
        end do

        call release(set)
      end if

    case('incompletegammaq')
      if (stringParameter('style') .eq. 'command') then
        aValue = realParameter('reals', 0)
        xValue = realParameter('reals', 1)
        call incompleteGammaQ(aValue, xValue, myAnswer, foreignStatus)
if (foreignStatus .ne. 0) call error('badArguments','')
        print *, 'answer = ', myAnswer

      else
        setTabName = stringParameter('tab')
        aColName = stringParameter('realcols', 0)
        xColName = stringParameter('realcols', 1)
        answerColName = stringParameter('answercol')

        call splitSetTabName(setTabName, setName, tabName)

        set = dataSet(setName, MODIFY)
        tab = table(set, tabName)
        aValues => real32Data(column(tab, aColName))
        xValues => real32Data(column(tab, xColName))

        if (hasColumn(tab, answerColName)) then
          call deleteColumn(tab, answerColName)
        end if
        answerValues => real32Data(addColumn(tab, answerColName, REAL32))

        do row = 0, numberOfRows(tab) - 1
          call incompleteGammaQ(aValues(row), xValues(row), answerValues(row)&
            , foreignStatus)
if (foreignStatus .ne. 0) call error('badArguments','')
        end do

        call release(set)
      end if

    case('sensitivity')
      if (stringParameter('style') .eq. 'command') then
        bkgValue = realParameter('reals', 0)
        likeValue = realParameter('reals', 1)
        call minDetPoissonCounts(bkgValue, likeValue, myAnswer&
          , myAnswerUncert, foreignStatus)

if (foreignStatus .ne. 0) call error('badArguments','')
        print *, 'answer = ', myAnswer, '+-', myAnswerUncert

      else
        setTabName = stringParameter('tab')
        bkgColName = stringParameter('realcols', 0)
        likeColName = stringParameter('realcols', 1)
        answerColName = stringParameter('answercol')

        call splitSetTabName(setTabName, setName, tabName)

        set = dataSet(setName, MODIFY)
        tab = table(set, tabName)
        bkgValues => real32Data(column(tab, bkgColName))
        likeValues => real32Data(column(tab, likeColName))

        if (hasColumn(tab, answerColName)) then
          call deleteColumn(tab, answerColName)
        end if
        answerValues => real32Data(addColumn(tab, answerColName, REAL32))

        do row = 0, numberOfRows(tab) - 1
          call minDetPoissonCounts(bkgValues(row), likeValues(row)&
            , answerValues(row), myAnswerUncert, status=foreignStatus)

if (foreignStatus .ne. 0) then
print *, 'row, bkg, like, status: ', row, bkgValues(row), likeValues(row), foreignStatus
if (foreignStatus < 0) call error('badArguments','')
end if
        end do

        call release(set)
      end if

    case default
call error('badFunction','')
    end select

  end subroutine poissontool

!  !---------------------------------------------------------------------
!  subroutine readParameters
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    !-------------------------------------------------------------------
!    ! Start
!
!  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module poissontool_mod
